.\" Emacs, this is in -*- nroff -*- mode "
.TH dirbrowser2 7 "14th September 1995" "Directory Browser 2.3" "Tk Command Extension"

.SH NAME
dirbrowser2 \- Motif-style select file from directory

.SH SYNOPSIS
.BI dirbrowser2 " pathName ?options?"

.SH OPTIONS
.SS BASIC
.BI -cancelexception
.RS
Generates an exception when the cancel button is pressed, instead of
returning an empty string.
.RE
.BI -defaultfile " fileName"
.RS
Sets the default file selected. If a path to the file is given, this
overrides the value in the
.B -directory
option.
.RE
.BI -directory " directoryName"
.RS
Sets the initial directory that the browser starts in. Defaults to
.I .
and is overridden by the path specified as part of the
.B -defaultfile
if a path is given there.
.RE
.BI -dodotfiles
.RS
Tells the file name matching engine to include files starting with a
"." in it's search. Normally these are excluded.
.RE
.BI -filemask " pattern"
.RS
Gives a pattern of files to match. Defaults to
.I *
.RE
.BI -globalgrab
.RS
Tells the browser to use a global grab as opposed to a local grab.
.RE
.BI -message " text"
.RS
Gives a string to be displayed to the user at the top of the window.
.RE
.BI -usecd
.RS
Tells the directory processor to use cd (in a sub-process) to perform
filename resolution for all directory names, and not just that of the
starting directory.
.RE
.SS ADVANCED
.BI -aftersetuphook " command"
.RS
Gives a command to be run after all the widgets and their bindings
within the browser have been created. This command will have the
pathname of the toplevel of the browser appended. This is the only way
to change the default bindings of the browser.
.RE
.BI -extendwidget " command"
.RS
Gives a commmand to be run immediately after the creation of the
toplevel widget, but before any operations have been performed on
it. The command will have the pathname of the toplevel of the browser
appended. This is the best way to add a new sub-panel to the top of
the browser, and if used, it is recommended that the
.B -message
option is not used.
.RE
.BI -extrabuttons " list"
.RS
Specifies a list of buttons to be added between the Cancel and the
Help buttons. Each element specifies one button and consists of a
sublist; the first element of which is the textual label on the
button, and the second element of which is the button command.
.RE
.BI -filetest " expression"
.RS
The expression is evaluated at the toplevel by the directory search
engine for every file found in order to determine if that file is
suitable for adding to the list of files to select. Directory names
are not verified using this mechanism. The name of the file being
considered is available in the (global) variable
.BR dirbrowser_filename .
The expression defaults to
.IR 1 " (true)."
.RE
.BI -hotpaths " varName"
.RS
If this option ispresent, by holding down Mouse Button 3, a menu of
`hot' directories should be popped up. Choosing one of the directories
from the menu will result in the browser moving into that directory.
Also available on the menu is an option to allow the user to add the
current directory to the end of the menu, and another option to allow
the user to edit the contents of the menu. The first directory on the
menu is always the user's home directory (as specified in
.IR env(HOME) ).
.RE

.SH DESCRIPTION
The
.B dirbrowser2
command creates a new top-level window (given by the
.I pathName
argument, or a child of that if it exists) and makes it into a
directory browser. The additional options, described above, may only
be specified on the command line, and they configure aspects of the
directory browser such as its starting directory and search pattern.

The directory browser offers the user two listboxes and two entries,
the top/left entry/listbox allowing the user to change directory and
the file search pattern, and the bottom/right entry/listbox allowing
the user to specify which file they want to select. It makes no
(lasting) changes to the current process's current work directory.
Pressing the
.I Tab
key while in the bottom entry (filename entry) will invoke filename
completion if what is already present uniquely determines at least the
next character, otherwise performing the standard
.IR Tab -traversal
behaviour.

At the bottom of the window created are several buttons:
.TP
.B O.K.
Accepts the value specified in the lower entry (activated by pressing
Return in that entry, or by Double-Clicking in the righthand listbox).
.TP
.B Filter
Rescans using the filename and directory specified in the upper entry
(activated by pressing Return in that entry or by Double-Clicking in
the lefthand listbox).
.TP
.B Cancel
Cancels the file selection action, returning an empty string
(optionally generating an exception, see the
.BR -cancelexception
option).
.TP
.B Help
Brings up a window giving help on directory browser.
.RE

Optionally (see the
.B -hotpaths
option) the user may hold down Mouse Button 3 to get a menu of
user-specified `hot' paths, selecting one of which causes the browser
to change directory to the one specified and scan for files matching
the current mask. Also on this menu are:
.TP
.B Add Current Directory
Adds the current directory to the end of the menu. Disabled if the
current directory is already present on the menu.
.TP
.B Edit Directory Hotlist
Brings up a dialog box which allows the user to edit their
hotlist. This is described below.
.TP
.B Home Directory
This entry is always present, and allows the user to change into their
home directory (as determined by the variable
.IR env(HOME) ).
.RE

The labels of all the other entries on the hotlist are the last
pathname component of the directory they represent.

.SS BINDINGS
The default Tk bindings for the constituent widgets are augmented by
the following bindings:
.IP [1]
The browser is broadly focus-follows-mouse, with the main default
focus targets being the two entries and the
.B Cancel
and
.B Help
buttons.
.IP [2]
The key
.IR Return ,
when pressed in the upper (filter) entry, causes the browser to read
the new directory and filter from the entry and use them to rebuild
the data in the listboxes and the lower (file) entry, as if the
.B Filter
button was pressed.
.IP [3]
The key
.IR Return ,
when pressed in the lower (file) entry, causes the browser to return
the value specified in that entry, as if the
.B O.K.
button was pressed. This destroys the directory browser.
.IP [4]
The key
.I Escape
is equivalent to pressing the
.B Cancel
button.
.IP [5]
The keys
.I Help
and
.I F1
are equivalent to pressing the
.B Help
button.
.IP [6]
A Button 1 Release in the righthand (file) listbox causes the file
selected there to become the file currently in the lower entry.
.IP [7]
A Double Button 1 in the lefthand (directory) listbox causes that
directory to become the current directory. The
.B "<< Parent >>"
entry is equivalent in this respect to
.I ..
.IP [8]
A Double Button 1 in the righthand (file) listbox causes the file
selected there to be returned as the result of the directory browser
operation, similar to if the
.B O.K.
button had been pressed.
.IP [9]
A Button 3 Press causes the `hot' directory menu to pop up, if this
has been enabled using the
.B -hotpaths
option. By default, this is not enabled.
.IP [10]
The key
.IR Tab ,
when pressed in the filename (bottommost) entry, invokes filename
completion, and traverses the window only if no (further) completion
is currently possible.

.SH DIRECTORY HOTLIST EDITOR
The directory list editor consists of a listbox containing the current
directory hotlist, and a collection of buttons on the right of it
which allow operations to be carried out on the hotlist. There are no
programmer configurable portions to the hotlist editor.

.SS BUTTON ACTIONS
.TP
.B Move Item Up
Moves the currently selected item up one place in the hotlist
(provided it is not the topmost entry).
.TP
.B Move Item Down
Moves the currently selected item down one place in the hotlist
(provided it is not the bottommost entry).
.TP
.B Insert Directory
Inserts a new entry (bringing up a dialog box to allow the user to
enter it's value) in the hotlist after the current entry.
.TP
.B Modify Entry
Brings up a dialog box which allows the user to modify the value of
the currently selected entry.
.TP
.B Delete Entry
Deletes the current entry from the hotlist.
.TP
.B O.K.
Accepts all changes made to the directory hotlist.
.TP
.B Cancel
Rejects all changes made to the directory hotlist.
.TP
.B Help
Brings up a dialog box with help on the directory hotlist editor.
.RE

.SS BINDINGS
The following bindings extend the editor beyond the Tk defaults:
.IP [1]
The key 
.I u
invokes the
.B Move Item Up
button.
.IP [2]
The key
.I d
invokes the
.B Move Item Down
button.
.IP [3]
The keys
.I i
and
.I Insert
invoke the
.B Insert Directory
button.
.IP [4]
The key
.I m
invokes the
.B Modify Entry
button.
.IP [5]
The keys
.I Backspace
and
.I Delete
invoke the
.B Delete Entry
button.
.IP [6]
The key
.I Return
invokes the
.B O.K.
button.
.IP [7]
The key
.I Escape
invokes the
.B Cancel
button.
.IP [8]
The keys
.I Help
and
.I F1
invoke the
.B Help
button.
.RE

.SH GLOBAL VARIABLES
.TP
.B dirbrowser
Array containing all private state to the directory browser. Only the
.B filterbody
and
.B filebody
elements (the contents of the upper and lower entries respectively)
are likely to be of much use to anyone tampering with this code. After
setting these either of these variables, it is advised that the
programmer call
.B dirbrowser_filterCommand
with no arguments to force a rescan using the updated values.
.TP
.B dirbrowser_filename
Variable containing the name of the file currently being considered in
the file test (see
.B -filetest
option). Never read by the filename globbing mechanism, so modifying
is pointless (and the precise effects are undefined).

.SH PRIVATE PROCEDURES
.SS INTERNAL
.TP
.B __dirbrowser_exit
Used to clean up.
.B "DO NOT CALL. EXIT THE APP USING `exit', NOT `destroy .'"

.SS DIRECTORY MANAGEMENT
.TP
.B dirbrowser_chdir
Like cd, but doesn't change the current process's idea of the
.I CWD.
.TP
.B dirbrowser_enddir
Shuts down directory subsystem (null-op currently).
.TP
.B dirbrowser_initdir
Initialises the directory subsystem (with initial dir).
.TP
.B dirbrowser_pwd
Companion pwd for
.BR dirbrowser_chdir .

.SS HOT DIRECTORY LIST MANAGEMENT
.TP
.B dirbrowser_do_menu
Handle menu for hot directories.
.TP
.B dirbrowser_edit_dirlist
Allows the user to edit their directory list.
.TP
.B dirbrowser_edit_*
Auxiliary functions for above.

.SS HIGHLIGHTING (PSEUDO- FOCUS FOLLOWS MOUSE)
.TP
.B dirbrowser_enterFilter
Action on entering the filter part of the dialog box.
.TP
.B dirbrowser_hl
Handles all changes in highlighting.
.TP
.B dirbrowser_leaveFilter
Action on leaving the filter part of the dialog box.
.TP
.B dirbrowser_resetFilter
In case things get confused (as _can_ happen).

.SS USER ACTION HANDLERS
.TP
.B dirbrowser_filterCommand
Double-click on directory /
.I Return
in filter box.
.TP
.B dirbrowser_helpCommand
Bring up a help dialog.
.TP
.B dirbrowser_complete
Performs filename completion.

.SS EXTENDED GLOBBING FACILITIES
.TP
.B dirbrowser_getdirlist
Gets a (sorted) list of subdirectories of the specified directory
(taking account of the
.B -dodotfiles
option).
.TP
.B dirbrowser_getfilelist
Gets a (sorted) list of files in the specified directory matching the
specified mask and satisfying the conditional expression supplied with
the
.B -filetest
option. Takes account of the
.B -dodotfiles
option.
.RE

.SH REQUIRES
Tcl7.4+, Tk4.0+,
.BR parseargs ", " recursivebind ", " with_grab_focus
.I (ie. dkflib.tcl)

.SH BUGS
I'm not happy with the grabbing and focussing mechanisms. In fact, I'd
probably class all uses of
.I recursivebind
as bugs in their own right.

Doesn't currently support the options database as a way to specify
options, and adding this will break current programs.

You can specify an already existant window as the
.I pathName
argument, and a directory browser will be created as a child of that
window, but no check is made to see if the window name that is finally
used is really in use, and to use a child of that if it is. I probably
ought to change this so that the name given must not previously exist,
but this would break several of my programs that use the old-style
syntax.  This is more commonly known as a M$Bug...

The filename completion conflicts with the keyboard traversal, and
doesn't work all well anyway.

It messes up badly with Tcl7.5/Tk4.1 under Windows due to the
existance of several roots to the directory tree, though this might be
fixed now (I can't test it there myself). It doesn't handle a ``//''
specified root either, so don't use that form...

If this works on the Mac, I'll be both surprised and pleased...

There is at least one bug in this manpage.

.SH TO DO
Fix (some of) the bugs.

Add drag-n-drop support.

.SH KEYWORDS
browse, dialog, directory, file, select

.SH SEE ALSO
.BR entry (n),
.BR file (n),
.BR glob (n),
.BR listbox (n),
.BR toplevel (n).

.SH AUTHOR
Donal K. Fellows: \fCfellowsd@cs.man.ac.uk\fR
