### -------------------------------------------------------------------------------------------------------
#
# PROCEDURE
#   dirbrowser2
#
# DESCRIPTION
#   Implements a directory browser, _without_ using cd in the current process...
#
# ARGUMENTS
#   parentwindow	 The path name of the parent.  NOT A KEYED ARGUMENT.	Defaults to  .
#			     If parentwindow is not an existsing widget, then it is assumed to be the name
#			     of the toplevel to create which contains the directory browser, so it's parent
#			     had better exist...
#   -defaultfile	 The default file to return. OVERRIDES -directory IF PATH SPECIFIED.
#										Defaults to  {}
#   -directory		 The directory which the search is to start in.		Defaults to  .
#   -filemask		 A glob to indicate which files to display.		Defaults to  *
#			     (all (non .*) directories are also shown)
#   -filetest		 A test to indicate whether the file in variable dirbrowser_filename should be
#			     included in the list. Evaluated at top-level.	Defaults to all files (1)
#   -message		 A message to the user to be displayed at the top of the browser.
#										Defaults to  {}
#   -dodotfiles		 A flag to indicate whether dot files should be included in the search.
#										(present ==  yes)
#   -cancelexception	 A flag to indicate whether Cancel causes an exception.	(present ==  yes)
#   -usecd		 A flag to indicate whether to use cd and pwd (in another process) to determine
#			     directory names. NB A separate process is required for this, so if the spawn
#			     fails, this setting will revert to no internally anyway.
#										(present ==  yes)
#   -hotpaths		 A variable containing a list of recently visited directories.
#										Defaults to  {} (no hotlist)
#   -extrabuttons	 List of lists {{text command} {text command} ...}	Defaults to  {}
#			     Each element will be translated into an extra button between Cancel and Help
#   -extendwidget	 Command to execute (with pathname of toplevel appended) before building any of the
#			     main body of the widgets. Using this right could take experimentation...
#										Defaults to  {}
#   -aftersetuphook	 Command to execute (with pathname of toplevel appended) after binding all of the
#			     main body of the widgets, but before scanning for files (directory resultion
#			     does occur before this though). Using this right could take experimentation...
#										Defaults to  {}
#   -globalgrab          A flag to indicate whether a global grab should be used.
#										(present ==  yes)
#
# RETURN VALUE
#   selected_filename	 if a file was selected
#   empty_string	 if `Cancel' was chosen (and -cancelexception wasn't present)
#
# ERRORS
#   none		 (NB: see parseargs)
#
# REQUIRES
#   ------ Tk 4.0 ------ ('Tis what I have, and it lets me do some things better than in 3.7 (version?))
#   parseargs		 option parser
#   recursivebind	 for performing a binding on a whole widget subtreee
#
# GLOBAL VARIALBES
#   dirbrowser_filename	   Use for conditions on acceptable files
#   dirbrowser(filterbody)  \				  / Actual current file filter
#   dirbrowser(toplevel)     |				 |  Window path of the toplevel of the listbox
#   dirbrowser(filebody)     |				 |  Actual current filename
#   dirbrowser(filetest)     |				 |  Test for acceptable files
#   dirbrowser(dodotfiles)   |				 |  Include .* files and directories?
#   dirbrowser(extrabut)     |				 |  Number of extra buttons added
#   dirbrowser(buttons)	     |				 |  Window path of buttons
#   dirbrowser(filter)	     |				 |  Window path of filter entry field
#   dirbrowser(fileref)	      \ For dirbrowser internal /   Window path of file entry field
#   dirbrowser(filtcnt)	      / use only. Do not modify \   Filter highlighting counter
#   dirbrowser(usecd)	     |				 |  Flag saying whether to use another process
#   dirbrowser(usesend)	     |				 |  Flag saying whether to use send to communicate
#   dirbrowser(dirlist)	     |				 |  Window path of directory listbox
#   dirbrowser(cwd)	     |				 |  `Current Work Directory' (when not using cd)
#   dirbrowser(proc)	     |				 |  Filehandle for subprocess	     \ When using
#   dirbrowser(sendid)	     |				 |  Interpreter name for subprocess  / chdir(2)
#   dirbrowser(editbutton)   |				 |  Internal to dirbrowser_edit_dirlist
#   dirbrowser(editcontents) |				 |     Ditto.
#   dirbrowser(dirlist*)     |				 |     Ditto.
#   dirbrowser(filelist)    /				  \ Window path of file listbox
#
# (SEMI-)PUBLIC ROUTINES
#   dirbrowser_enterFilter    \		 / Action on entering the filter part of the dialog box
#   dirbrowser_leaveFilter     \ Caveat /  Action on leaving the filter part of the dialog box
#   dirbrowser_resetFilter      >emptor<   In case things get confused (as _can_ happen)
#   dirbrowser_filterCommand   /	\  Double-click on directory / <Return> in filter box
#   dirbrowser_complete       /          \ <Tab> in filename. Returns whether a change was made
#   dirbrowser_hl	      Handles all changes in highlighting
#   dirbrowser_getdirlist     Gets list of subdirs in given dir (respecting flags)
#   dirbrowser_getfilelist    Gets list of files matching pattern in dir (respecting flags and expression)
#   dirbrowser_initdir	      Initialises the directory subsystem (with inital dir)
#   dirbrowser_chdir	      Like cd, but doesn't change the current process's idea of the CWD
#   dirbrowser_pwd	      Pwd for above chdir
#   dirbrowser_enddir	      Shuts down directory subsystem (null-op currently)
#   dirbrowser_do_menu	      Handle menu for hot directories
#   dirbrowser_edit_dirlist   Allows the user to edit their directory list
#   dirbrowser_edit_*	      Auxiliary functions for above.
#   __dirbrowser_exit	      Used to clean up. DO NOT CALL. EXIT THE APP USING `exit', NOT `destroy .'
#
# BUGS
#   Probably lots (see public routines :) but I'm not aware of any major ones  :)
#   Maybe I should put more `catch'es in...
#
# TO DO
#   1) Reprogram to use option database instead of hardcoded fonts, etc.  
#	     DONE!
#   2) Adjust to use a separate process to track the current directory
#	     MADE AN OPTION, SINCE IT CHANGES THE BEHAVIOUR WITH SYMLINKS SOMETHING ROTTEN...
#   3) Write a cleverer (i.e. atomic) semaphore. This will need work (maybe even assembler )-:
#	     NOT NEEDED!
#   4) Update to work with tk4.0
#	     DONE (I think)
#   5) Write a MAN page (there is too much documentation in this file... :^)
#            MOSTLY DONE
#   6) Add filename completion
#            DONE (I think)
#

proc dirbrowser2 args {
    global dirbrowser_filename   \
	   dirbrowser
    upvar #0  dirbrowser(buttons)   buttons  \
	      dirbrowser(filter)    filter   \
	      dirbrowser(fileref)   fileref  \
	      dirbrowser(filtcnt)   count    \
              dirbrowser(dirlist)   dirlist  \
              dirbrowser(filelist)  filelist \
	      dirbrowser(toplevel)  w        \
	      dirbrowser(extrabut)  numXbut

    set count 0

    set allowed_switches(-cancelexception) 0
    set allowed_switches(-dodotfiles)      0
    set allowed_switches(-globalgrab)      0
    set allowed_switches(-usecd)           0
    set allowed_switches(-aftersetuphook)  1
    set allowed_switches(-defaultfile)     1
    set allowed_switches(-extendwidget)    1
    set allowed_switches(-extrabuttons)    1
    set allowed_switches(-hotpaths)        1
    set allowed_switches(-message)         1
    set allowed_switches(-directory)       2
    set allowed_switches(-filemask)        2
    set allowed_switches(-filetest)        2
    set switchval(-directory)              "."
    set switchval(-filemask)               "*"
    set switchval(-filetest)               "1"

    set parentwindow [parseargs allowed_switches switchval $args]
    
    set dirbrowser(usecd) $switchval(-usecd)
    onexit {
	if [string length $switchval(-defaultfile)] {
	    set filedir [file dirname $switchval(-defaultfile)]
	    set filename [file tail $switchval(-defaultfile)]
	    if [regexp {^[~/]} $filedir] {
		dirbrowser_initdir $filedir
	    } else {
		dirbrowser_initdir .
		dirbrowser_chdir $filedir
	    }
	    # Beware of symlinks to directories!
	    set filestats(type) {}
	    catch {
		file stat $filename filestats
	    }
	    if {$filestats(type) == "directory"} {
		dirbrowser_chdir $filename
		set switchval(-defaultfile) {}
	    } else {
		set switchval(-defaultfile) $filename
	    }
	} else {
	    dirbrowser_initdir .
	    dirbrowser_chdir $switchval(-directory)
	}
	set switchval(-directory) [dirbrowser_pwd]

	set dirbrowser(filetest)    $switchval(-filetest)
	set dirbrowser(dodotfiles)  $switchval(-dodotfiles)
	set dirbrowser(filterbody)  $switchval(-directory)/$switchval(-filemask)
	
	update idletasks

	if [winfo exists $parentwindow] {
	    if {[string length $parentwindow] && [string compare $parentwindow "."]} {
		set w ${parentwindow}.dirbrowser
	    } else {
		set parentwindow .
		set w .dirbrowser
	    }
	    toplevel $w -class DKFDirbrowser
	} else {
	    set w $parentwindow				;# Really me!
	    toplevel $w -class DKFDirbrowser
	    set parentwindow [winfo parent $w]		;# My real parent
	}

	if [llength $switchval(-extendwidget)] {
	    eval $switchval(-extendwidget) $w
	}

	wm transient $w $parentwindow
	wm group $w $parentwindow

	# References to the principal interactive widgets/widget groups ----------------------------------

	set buttons   $w.buttons
	set filter    $w.filter.body.text
	set fileref   $w.filename.body.text
	set dirlist   $w.dirs.body.2.list
	set filelist  $w.files.body.2.list
	set numXbut   [llength $switchval(-extrabuttons)]

	# The frames and labels (ie. the structure) ------------------------------------------------------

	if [string length $switchval(-message)] {
	    message $w.message  -justify left  -relief raised  -bd 2  -text $switchval(-message)  -width 12c
	    pack $w.message  -side top  -fill x  -expand 1  -anchor w
	}

	frame $w.filter			-relief raised  -bd 2
	frame $w.filename		-relief raised  -bd 2
	frame $w.dirs			-relief raised  -bd 2  -width 7c
	frame $w.files			-relief raised  -bd 2  -width 7c
	frame $w.buttons		-relief raised  -bd 2
	frame $w.buttons.okbrd		-relief sunken  -bd 1
	frame $w.buttons.filtbrd	-relief flat    -bd 1
	frame $w.buttons.canbrd		-relief flat    -bd 1
	for {set i 0} {$i<$numXbut} {incr i} {
	    frame $w.buttons.butbrd$i	-relief flat	-bd 1
	}
	frame $w.buttons.hlpbrd		-relief flat    -bd 1
	frame $w.filter.body		-relief groove  -bd 4
	frame $w.filename.body		-relief groove  -bd 4
	frame $w.dirs.body
	frame $w.dirs.body.1
	frame $w.dirs.body.1.spc
	frame $w.dirs.body.2
	frame $w.files.body
	frame $w.files.body.1
	frame $w.files.body.1.spc
	frame $w.files.body.2
	label $w.filter.label		-text "Filter:"
	label $w.filename.label		-text "Filename:"
	label $w.dirs.label		-text "Directory"  -anchor e
	label $w.files.label		-text "Files"  -anchor e

	#	What		Where		Expand?		Fill?		Other bits
	pack  $w.filter		-side top	-expand 1	-fill x
	pack  $w.buttons	-side bottom	-expand 1	-fill x
	pack  $w.filename	-side bottom	-expand 1	-fill x
	pack  $w.dirs		-side left	-expand 1
	pack  $w.files		-side left	-expand 1
	pack  $w.buttons.okbrd	-side left	-expand 1			-padx 2m	-pady 2m
	pack  $w.buttons.filtbrd -side left	-expand 1			-padx 2m	-pady 2m
	pack  $w.buttons.canbrd	-side left	-expand 1			-padx 2m	-pady 2m
	for {set i 0} {$i<$numXbut} {incr i} {
	    pack  $w.buttons.butbrd$i -side left -expand 1			-padx 2m	-pady 2m
	}
	pack  $w.buttons.hlpbrd	-side left	-expand 1			-padx 2m	-pady 2m
	pack  $w.filter.label	-side left
	pack  $w.filter.body	-side right					-padx 1m	-pady 1m
	pack  $w.filename.label	-side left
	pack  $w.filename.body	-side right					-padx 1m	-pady 1m
	pack  $w.dirs.label	-side top
	pack  $w.dirs.body	-side top	-expand 1	-fill x		-anchor n
	pack  $w.dirs.body.2	-side left			-fill y
	pack  $w.dirs.body.1	-side left			-fill y
	pack  $w.dirs.body.1.spc -side bottom	-expand 0
	pack  $w.files.label	-side top
	pack  $w.files.body	-side top	-expand 1	-fill x		-anchor n
	pack  $w.files.body.2	-side left			-fill y
	pack  $w.files.body.1	-side left			-fill y
	pack  $w.files.body.1.spc -side bottom	-expand 0

	# The buttons at the bottom of the window --------------------------------------------------------

	button $buttons.ok      -text "O.K."    -command "destroy $w"
	button $buttons.cancel  -text "Cancel"  -command "set dirbrowser_filename {}; destroy $w"
	button $buttons.filter  -text "Filter"  -command dirbrowser_filterCommand
	button $buttons.help    -text "Help"    -command [list dirbrowser_helpCommand $w]
	
	pack $buttons.ok      -in $buttons.okbrd	-side left -padx 3 -pady 3 -ipadx 3 -ipady 3
	pack $buttons.filter  -in $buttons.filtbrd	-side left -padx 3 -pady 3 -ipadx 3 -ipady 3
	pack $buttons.cancel  -in $buttons.canbrd	-side left -padx 3 -pady 3 -ipadx 3 -ipady 3
	pack $buttons.help    -in $buttons.hlpbrd	-side left -padx 3 -pady 3 -ipadx 3 -ipady 3

	for {set i 0} {$i<$numXbut} {incr i} {
	    set thisbut [lindex $switchval(-extrabuttons) $i]
	    pack [button $buttons.butbrd$i.but -text [lindex $thisbut 0] -command [lindex $thisbut 1]] \
		    -side left -padx 3 -pady 3 -ipadx 3 -ipady 3
	    bind $buttons.butbrd$i.but <Enter> [list dirbrowser_hl $i]
	}

	bind $buttons.ok      <Enter> {dirbrowser_hl ok}
	bind $buttons.okbrd   <Enter> {dirbrowser_hl ok}
	bind $buttons.filter  <Enter> {dirbrowser_hl filter}
	bind $buttons.filtbrd <Enter> {dirbrowser_hl filter}
	bind $buttons.cancel  <Enter> {dirbrowser_hl cancel}
	bind $buttons.canbrd  <Enter> {dirbrowser_hl cancel}
	bind $buttons.help    <Enter> {dirbrowser_hl help}
	bind $buttons.hlpbrd  <Enter> {dirbrowser_hl help}
	bind $buttons         <Leave> {dirbrowser_hl ok}

	# The entry area at the top ----------------------------------------------------------------------

	entry     $filter              -textvariable dirbrowser(filterbody) \
		-xscrollcommand "$w.filter.body.scrl set"
	scrollbar $w.filter.body.scrl  -command "$filter xview"  -orient horizontal  -takefocus 0

	pack $w.filter.body.scrl  -side bottom  -fill x
	pack $filter

	recursivebind $w.filter -bindself 1  <Enter>  {dirbrowser_hl filter}
	recursivebind $w.filter -bindself 1  <Leave>  {dirbrowser_hl ok}
	bind $w.filter <Motion> {dirbrowser_hl filter}

	# The entry area at the bottom -------------------------------------------------------------------

	entry $fileref           -textvariable dirbrowser(filebody) \
		-xscrollcommand "$w.filename.body.scrl set"
	scrollbar $w.filename.body.scrl  -command "$fileref xview"  -orient horizontal  -takefocus 0
	set dirbrowser(filebody) "$switchval(-directory)/$switchval(-defaultfile)"
	
	pack $w.filename.body.scrl  -side bottom  -fill x
	pack $fileref

	# The directory list at the side -----------------------------------------------------------------

	scrollbar  $w.dirsYScroll	-command "$dirlist yview"  -takefocus 0
	scrollbar  $w.dirsXScroll	-command "$dirlist xview"  -orient horizontal  -takefocus 0
	listbox    $dirlist		-relief raised  -bd 2   -takefocus 0	\
		-yscrollcommand "$w.dirsYScroll set" -xscrollcommand "$w.dirsXScroll set"

	pack $w.dirsXScroll  -in $w.dirs.body.2  -side bottom  -fill x
	pack $w.dirsYScroll  -in $w.dirs.body.1  -fill y -expand 1
	pack $dirlist        -side left
	$w.dirs.body.1.spc configure -height [expr	\
		[  $w.dirsXScroll cget -width] +	\
		2*[$w.dirsXScroll cget -bd]    +	\
		2*[$w.dirsXScroll cget -highlightthickness]]

	bind          $dirlist             <Double-Button-1> {dirbrowser_filterCommand [selection get]}
	recursivebind $w.dirs -bindself 1  <Enter>           dirbrowser_enterFilter
	recursivebind $w.dirs -bindself 1  <Leave>           dirbrowser_leaveFilter
	bind $w.dirsXScroll <Enter> dirbrowser_enterFilter
	bind $w.dirsYScroll <Enter> dirbrowser_enterFilter
	bind $w.dirsXScroll <Leave> dirbrowser_leaveFilter
	bind $w.dirsYScroll <Leave> dirbrowser_leaveFilter

	# The file list at the other side ----------------------------------------------------------------
	
	scrollbar  $w.filesYScroll  -command "$filelist yview"  -takefocus 0
	scrollbar  $w.filesXScroll  -command "$filelist xview"  -orient horizontal  -takefocus 0
	listbox    $filelist -relief raised  -bd 2  -takefocus 0	\
		-yscrollcommand "$w.filesYScroll set" -xscrollcommand "$w.filesXScroll set"

	set dirbrowser_filename $switchval(-directory)

	pack $w.filesXScroll  -in $w.files.body.2  -side bottom  -fill x
	pack $w.filesYScroll  -in $w.files.body.1  -fill y -expand 1
	pack $filelist             -side left
	$w.files.body.1.spc configure -height [expr	\
		[  $w.filesXScroll cget -width] +	\
		2*[$w.filesXScroll cget -bd]    +	\
		2*[$w.filesXScroll cget -highlightthickness]]

	bind $filelist <ButtonRelease-1> {
	    set dirbrowser_filename [selection get]
	    regsub {/[^/]*$} $dirbrowser(filebody) "/$dirbrowser_filename" dirbrowser(filebody)
	    tkEntrySeeInsert $dirbrowser(fileref)
	}
	bind $filelist <Double-Button-1> {
	    set dirbrowser_filename [selection get]
	    $dirbrowser(buttons).ok invoke
	    break; # How stupid can a bug be? The above line deletes the widget, but without this line
	    #      # there are still bindings to be applied... This is the quickest fix.
	}

	# Only return when the user has supplied a value -------------------------------------------------

	bind $w       <Escape>  [list $buttons.cancel invoke]
	bind $fileref <Return>  [list $buttons.ok invoke]
	bind $filter  <Return>  dirbrowser_filterCommand
	if [llength $switchval(-hotpaths)] {
	    bind $w <3> [list dirbrowser_do_menu $w $switchval(-hotpaths) %X %Y]
	}
	bind $w <F1> [list dirbrowser_helpCommand $w]
	catch {
	    bind $w <Help> [list dirbrowser_helpCommand $w]
	}
	bind $fileref <Tab> {
	    if [dirbrowser_complete] {
		dirbrowser_filterCommand
		break
	    } else {
		dirbrowser_filterCommand
	    }
	}

	if [llength $switchval(-aftersetuphook)] {
	    eval $switchval(-aftersetuphook) $w
	}

	dirbrowser_filterCommand .

	wm withdraw $w
	update idletasks
	wm title $w "Directory Browser"
	catch {
	    regexp {([-+][0-9]+)([-+][0-9]+)} [wm geometry [wm toplevel $parentwindow]] geom geomX geomY
	    wm geometry $w +[expr 30 $geomX]+[expr 30 $geomY]
	}
	wm deiconify $w

	tkwait visibility $w
	if $switchval(-globalgrab) {
	    with_grab_focus $w $fileref -global "tkwait window $w"
	} else {
	    with_grab_focus $w $fileref         "tkwait window $w"
	}
    } {
	dirbrowser_enddir
    }
    if [string length $dirbrowser_filename] {
	return $dirbrowser(filebody)
    } elseif $switchval(-cancelexception) {
	return -code 64 "File Selection Cancelled"
    } else {
	return {}
    }
}

# --- BEGINNING OF PRIVATE PROCEDURES --------------------------------------------------------------------

### ----------------------------------------------------------------------------------------------------
# Highlighting
;proc dirbrowser_hl which {
    global dirbrowser

    switch $which {
	ok {
	    $dirbrowser(buttons).okbrd   configure -relief sunken
	    $dirbrowser(buttons).filtbrd configure -relief flat
	    $dirbrowser(buttons).canbrd  configure -relief flat
	    $dirbrowser(buttons).hlpbrd  configure -relief flat
	    focus $dirbrowser(fileref)
	}
	filter {
	    $dirbrowser(buttons).okbrd   configure -relief flat
	    $dirbrowser(buttons).filtbrd configure -relief sunken
	    $dirbrowser(buttons).canbrd  configure -relief flat
	    $dirbrowser(buttons).hlpbrd  configure -relief flat
	    focus $dirbrowser(filter)
	}
	cancel {
	    $dirbrowser(buttons).okbrd   configure -relief flat
	    $dirbrowser(buttons).filtbrd configure -relief flat
	    $dirbrowser(buttons).canbrd  configure -relief sunken
	    $dirbrowser(buttons).hlpbrd  configure -relief flat
	    focus $dirbrowser(buttons).cancel
	}
	help {
	    $dirbrowser(buttons).okbrd   configure -relief flat
	    $dirbrowser(buttons).filtbrd configure -relief flat
	    $dirbrowser(buttons).canbrd  configure -relief flat
	    $dirbrowser(buttons).hlpbrd  configure -relief sunken
	    focus $dirbrowser(buttons).help
	}
	default {
	    $dirbrowser(buttons).okbrd   configure -relief flat
	    $dirbrowser(buttons).filtbrd configure -relief flat
	    $dirbrowser(buttons).canbrd  configure -relief flat
	    $dirbrowser(buttons).hlpbrd  configure -relief flat
	    for {set i 0} {$i<$dirbrowser(extrabut)} {incr i} {
		if {$i==$which} {
		    $dirbrowser(buttons).butbrd$i configure -relief sunken
		} else {
		    $dirbrowser(buttons).butbrd$i configure -relief flat
		}
	    }
	    focus $dirbrowser(buttons).butbrd$which.but
	    return
	}
    }
    for {set i 0} {$i<$dirbrowser(extrabut)} {incr i} {
	$dirbrowser(buttons).butbrd$i configure -relief flat
    }
}

# Atomicity bug should now be fixed.
#
# Oh, it isn't really (for various reasons). Fixed by a wholly different hack which makes usage over slow
# links an amusing concept... :)
#
# OOer! 'Tis worse than an atomicity bug since each handler doesn't get preempted by another anyway. What
# is happening is even more complicated. I'll have fun sorting this out (but I suspect that it is due to
# multiple bindings being applied to a single event <sigh>

# Increments the flag indicating if in the filter part, changing the focus(etc.) if I'm in for real!
;proc dirbrowser_enterFilter {} {
    global dirbrowser				;# DEBUG_COUNT ;# not needed now..

    incr dirbrowser(filtcnt)
    if {$dirbrowser(filtcnt) > 0} {
	dirbrowser_hl filter
    }
}

# Decrements the flag indicating if in the filter part, changing the focus(etc.) if I'm out for real!
;proc dirbrowser_leaveFilter {} {
    global dirbrowser				;# DEBUG_COUNT ;# not needed now..

    incr dirbrowser(filtcnt) -1
    if {$dirbrowser(filtcnt) <= 0} {
	set dirbrowser(filtcnt) 0
	dirbrowser_hl ok
    }
}

# Resets the flag indicating if in the filter part, changing the focus(etc.)
;proc dirbrowser_resetFilter {} {
    global dirbrowser				;# DEBUG_COUNT ;# not needed now..

    set dirbrowser(filtcnt) 0
    dirbrowser_hl ok
}    


### ----------------------------------------------------------------------------------------------------
# Major directory browser actions

;proc dirbrowser_getdirlist {directory} {
    global dirbrowser
    if [catch {set d1 [glob $directory/*/]}] {
	set d1 {}
    }
    if {!$dirbrowser(dodotfiles) || [catch {set d2 [glob $directory/.*/]}]} {
	set d2 {}
    }
    return [lsort [concat $d1 $d2]]
}

;proc dirbrowser_getfilelist {directory filemask} {
    global dirbrowser dirbrowser_filename
    set files {}
    catch {
	foreach dirbrowserFilename [glob $directory/$filemask] {
	    file stat $dirbrowserFilename filestats
	    set isadir [expr $filestats(mode) & 040000]
	    set dirbrowser_filename $dirbrowserFilename
	    if {!$isadir && [uplevel #0 expr $dirbrowser(filetest)]} {
		lappend files $dirbrowserFilename
	    }
	}
    }
    if !$dirbrowser(dodotfiles) {
	catch {
	    foreach dirbrowserFilename [glob $directory/.$filemask] {
		file stat $dirbrowserFilename filestats
		set isadir [expr $filestats(mode) & 040000]
		set dirbrowser_filename $dirbrowserFilename
		if {!$isadir && [uplevel #0 expr $dirbrowser(filetest)]} {
		    lappend files $dirbrowserFilename
		}
	    }
	}
    }
    return [lsort $files]
}

;proc dirbrowser_complete {} {
    global dirbrowser

    set mask [file tail $dirbrowser(filterbody)]
    set startingpoint [file tail $dirbrowser(filebody)]
    set directory [file dirname $dirbrowser(filebody)]

    foreach d [dirbrowser_getdirlist $directory] {
	if ![string first $dirbrowser(filebody) $d] {
	    lappend expands $d
	}
    }
    foreach f [dirbrowser_getfilelist $directory $mask] {
	if ![string first $dirbrowser(filebody) $f] {
	    lappend expands $f
	}
    }
    switch [llength $expands] {
	0 {
	    return 0
	}
	1 {
	    set dirbrowser(filebody) [lindex $expands 0]
	    set dirbrowser(filterbody) "[file dirname $dirbrowser(filebody)]/$mask"
	    return 1
	}
	default {
	    set flag 0
	    set successflag 0
	    set matchedflag 0
	    set mlen [string length $dirbrowser(filebody)]
	    while 1 {
		set match [string range [lindex $expands 0] 0 $mlen]
		foreach d $expands {
		    if [string first $match $d] {
			set flag 1
		    } elseif ![string compare $match $d] {
			set matchedflag 1
		    }
		}
		if $flag {
		    # No longer prefix
		    return $successflag
		} else {
		    set dirbrowser(filebody) $match
		    set dirbrowser(filterbody) "[file dirname $dirbrowser(filebody)]/$mask"
		    if $matchedflag {
			return 1
		    }
		    set successflag 1
		    incr mlen
		}
	    }
	}
    }
}

# Code to update the contents of the various buffers when the filtering changes.
# This is the real key to this whole file!
;proc dirbrowser_filterCommand {{dirname {}}} {
    global dirbrowser

    if {![string compare $dirname "<< Parent >>"]} {
	dirbrowser_chdir ..
    } elseif {[string compare $dirname {}] && $dirname != "."} {
	dirbrowser_chdir $dirname
    } else {
	dirbrowser_chdir [file dirname $dirbrowser(filterbody)]
    }

    set directory [dirbrowser_pwd]
    set filemask [file tail $dirbrowser(filterbody)]
    set dirbrowser(filterbody) "$directory/$filemask"
    set dirbrowser(filebody) "$directory/[file tail $dirbrowser(filebody)]"

    $dirbrowser(dirlist) delete 0 end
    if {![regexp {^/*$} $directory]} {$dirbrowser(dirlist) insert end "<< Parent >>"}
    foreach dir [dirbrowser_getdirlist $directory] {
	regexp {/([^/]*)/$} $dir foo dir
	if {!($dir == ".") && !($dir == "..")} {
	    $dirbrowser(dirlist) insert end $dir
	}
    }

    $dirbrowser(filelist) delete 0 end
    foreach file [dirbrowser_getfilelist $directory $filemask] {
	$dirbrowser(filelist) insert end [file tail $file]
    }
    catch {
	tkEntrySeeInsert $dirbrowser(filter)
	tkEntrySeeInsert $dirbrowser(fileref)
    }
}

;proc dirbrowser_helpCommand {w} {
    tk_dialog $w.help "Directory Browser Help" "The directory browser is split into two main regions:\
	    \n    the Filter/Directory Area, and\n    the File area.\n\n  \
	    In the Filter Area, you can type in the name of the directory you wish to change to and\
	    the filename filter you wish to apply there. Pressing Return (or the Filter button) causes\
	    the browser to use the values you have just entered. Double-clicking on one of the directory\
	    names listed in the list at the side will cause the system to change into that directory.\n\n  \
	    In the File Area, you may type in the name of the file that you want (pressing Return or the\
	    O.K. button to accept what you have typed in). Double-clicking on one of the listed files will\
	    cause that to be selected as the required file, and single clicking will transfer that file to\
	    the filename entry box.\n\n  \
	    Pressing Escape anywhere cancels, and pressing the right mouse button brings up a hot directory\
	    list (if one is available). This help may be brought up by pressing F1 or Help." \
	    info 0 {O.K.}    
}

### ----------------------------------------------------------------------------------------------------
# Improving on plain cd/pwd
;proc dirbrowser_initdir d {
    global dirbrowser
    set cmd [list cd $d]
    if $dirbrowser(usecd) {
	if [llength [array names dirbrowser proc]] {
	    if [catch {
		send $dirbrowser(sendid) $cmd
		set dirbrowser(usesend) 1
	    }] {
		puts $dirbrowser(proc) $cmd
		flush $dirbrowser(proc)
		set dirbrowser(usesend) 0
	    }
	} elseif {[catch {
	    if [catch {
		set dirbrowser(proc) [open "|[info nameofexecutable]" r+]
	    }] {
		set dirbrowser(proc) [open "|wish4.0" r+]
	    }
	    puts $dirbrowser(proc) {wm withdraw .}
	    puts $dirbrowser(proc) {puts stdout [tk appname]}
	    puts $dirbrowser(proc) {flush stdout}
	    flush $dirbrowser(proc)
	    # Kill at the appropriate time
	    rename exit __dirbrowser_exit
	    ;proc exit args {
		# The semicolon is to fool auto_mkindex
		global dirbrowser
		puts  $dirbrowser(proc) {exit}
		flush $dirbrowser(proc)
		close $dirbrowser(proc)
		eval __dirbrowser_exit $args
	    }
	    set dirbrowser(sendid) [gets $dirbrowser(proc)]
	    if [catch {
		send $dirbrowser(sendid) $cmd
		set dirbrowser(usesend) 1
	    }] {
		puts $dirbrowser(proc) $cmd
		flush $dirbrowser(proc)
		set dirbrowser(usesend) 0
	    }
	}]} {
	    set dirbrowser(usecd) 0
	}
    }
    if !$dirbrowser(usecd) {
	# This is not an else clause, since the above if can reset the flag
	set olddir [pwd]
	catch {
	    cd $d
	}
	set dirbrowser(cwd) "[pwd]/"
	cd $olddir
    }
}
;proc dirbrowser_chdir d {
    global dirbrowser

    if $dirbrowser(usecd) {
	if $dirbrowser(usesend) {
	    send $dirbrowser(sendid) [list cd $d]
	} else {
	    puts $dirbrowser(proc) [list cd $d]
	    flush $dirbrowser(proc)
	}
    } else {
	# dirbrowser(cwd) is the current directory, with a trailing slash
	switch $d {
	    . {
		# This space intentionally left blank
	    }
	    .. {
		set dirbrowser(cwd) [file dirname \
			[file dirname $dirbrowser(cwd)]]/
		### Chop off an element if not at the root
		# if {"$dirbrowser(cwd)" != "/"} {
		#   regexp {^(.*/)[^/]*/$} $dirbrowser(cwd) foo dirbrowser(cwd)
		# }
	    }
	    default {
		# Go there!
		if {[info tclversion]==7.4} {
		    if {[string index $d 0] == "/"} {
			set dirbrowser(cwd) "$d/"
		    } else {
			append dirbrowser(cwd) "$d/"
		    }
		} else {
		    global tcl_platform
		    switch $tcl_platform(platform) {
			windows {
			    if [string match ?:* $d] {
				set dirbrowser(cwd) "$d/"
			    } elseif {[string index $d 0] == "/"} {
				set dirbrowser(cwd) \
					"[string index $dirbrowser(cwd) 0]:$d/"
			    } else {
				append dirbrowser(cwd) "$d/"
			    }

			}
			default {
			    if {[string index $d 0] == "/"} {
				set dirbrowser(cwd) "$d/"
			    } else {
				append dirbrowser(cwd) "$d/"
			    }
			}
		    }
		}
	    }
	}
    }
}
;proc dirbrowser_pwd {} {
    global dirbrowser

    if $dirbrowser(usecd) {
	if $dirbrowser(usesend) {
	    return [send $dirbrowser(sendid) pwd]
	} else {
	    puts $dirbrowser(proc) {puts stdout [pwd]; flush stdout}
	    flush $dirbrowser(proc)
	    return [gets $dirbrowser(proc)]
	}
    } else {
	regexp {^((/*[^/]+)*)/+$} $dirbrowser(cwd) foo cwd
	return $cwd
    }
}
;proc dirbrowser_enddir {} {
    # Currently blank, as I'll leave the subprocess running for efficiency of future runs of dirbrowser
}

### ----------------------------------------------------------------------------------------------------
# Hot directory list management
;proc dirbrowser_do_menu {w dirlist x y} {
    global env
    upvar #0 $dirlist dl
    set cwd [dirbrowser_pwd]
    if [winfo exists $w.dirmenu] {
	$w.dirmenu delete 0 end
    } else {
	menu $w.dirmenu -tearoff 0
    }
    if {[lsearch -exact $dl $cwd]==-1} {
	$w.dirmenu add command -label "Add Current Directory" -command \
		[list lappend $dirlist $cwd]
    } else {
	$w.dirmenu add command -label "Add Current Directory" -state disabled
    }
    $w.dirmenu add command -label "Edit Directory Hotlist" -command \
	    [list dirbrowser_edit_dirlist $w $dirlist]
    $w.dirmenu add separator
    $w.dirmenu add command -label "Home Directory" -command [list dirbrowser_filterCommand $env(HOME)]
    foreach dir $dl {
	regexp {([^/]+|/)$} $dir shortdir
	$w.dirmenu add command -label $shortdir -command [list dirbrowser_filterCommand $dir]
    }
    tk_popup $w.dirmenu $x $y
}

;proc dirbrowser_edit_dirlist {w dirlist} {
    global dirbrowser
    upvar #0 $dirlist dl

    set p $w.dirlistedit
    toplevel $p -class DKFDirListEditor
    pack [frame $p.l] -side left -expand 1 -fill both
    pack [frame $p.r] -side left -fill both
    pack [listbox $p.l.l -yscrollcommand [list $p.l.s set] -selectmode single -exportselection 0] \
	    -side left -fill both -expand 1
    pack [scrollbar $p.l.s -command [list $p.l.l yview]] -side left -fill y

    pack [button $p.r.up -text "Move Item Up"	-command [list dirbrowser_edit_up     $p] \
	    -underline 10] -fill x
    pack [button $p.r.down -text "Move Item Down" -command [list dirbrowser_edit_down $p] \
	    -underline 10] -fill x
    pack [button $p.r.ins -text "Insert Directory" -command [list dirbrowser_edit_insert $p] \
	    -underline 0] -fill x
    pack [button $p.r.mod -text "Modify Entry"	-command [list dirbrowser_edit_modify $p] \
	    -underline 0] -fill x
    pack [button $p.r.del -text "Delete Entry"	-command [list dirbrowser_edit_delete $p]] -fill x
    pack [button $p.r.hlp -text "Help"		-command [list dirbrowser_edit_help   $p]] \
	    -fill x -side bottom
    pack [button $p.r.can -text "Cancel"	-command [list dirbrowser_edit_cancel $p]] \
	    -fill x -side bottom
    pack [button $p.r.ok -text "O.K."		-command [list dirbrowser_edit_ok     $p]] \
	    -fill x -side bottom

    foreach dir $dl {
	$p.l.l insert end $dir
    }
    if [llength $dl] {
	$p.l.l selection set 0
	$p.l.l see 0
    }

    bind $p u [list $p.r.up invoke]
    bind $p d [list $p.r.down invoke]
    catch {
	bind $p <Insert> [list $p.r.ins invoke]
    }
    bind $p i [list $p.r.ins invoke]
    bind $p m [list $p.r.mod invoke]
    catch {
	bind $p <Delete> [list $p.r.del invoke]
    }
    catch {
	bind $p <BackSpace> [list $p.r.del invoke]
    }
    bind $p <Return> [list $p.r.ok invoke]
    bind $p <Escape> [list $p.r.can invoke]
    bind $p <F1> [list $p.r.hlp invoke]
    catch {
	bind $p <Help> [list $p.r.hlp invoke]
    }

    wm withdraw $p
    wm title $p "Directory List Editor"
    if [llength [wm transient [winfo toplevel $w]]] {
	wm transient $p [wm transient [winfo toplevel $w]]
    } else {
	wm transient $p [winfo toplevel $w]
    }
    if [llength [wm group [winfo toplevel $w]]] {
	wm group $p [wm group [winfo toplevel $w]]
    }
    wm resizable $p 0 0
    catch {
	regexp {([-+][0-9]+)([-+][0-9]+)} [wm geometry [winfo toplevel $w]] geom geomX geomY
	wm geometry $p +[expr 30 $geomX]+[expr 30 $geomY]
    }
    wm deiconify $p

    with_grab_focus $p $p.l.l "tkwait window $p"
    if $dirbrowser(editbutton) {
	set dl $dirbrowser(editcontents)
    }
}

;proc dirbrowser_edit_entry {w default} {
    global dirbrowser
    set dirbrowser(dirlisteditentry) $default

    toplevel $w.entry -class DKFDirListEditorEntry
    pack [entry $w.entry.e -textvariable dirbrowser(dirlisteditentry)] -fill x
    pack [button $w.entry.ok -text O.K. -command [list dirbrowser_edit_entry_but $w.entry 0]] \
	    -expand 1 -side left -fill x
    pack [button $w.entry.can -text Cancel -command [list dirbrowser_edit_entry_but $w.entry 1]] \
	    -expand 1 -side left -fill x

    wm withdraw $w.entry
    if [llength [wm transient [winfo toplevel $w]]] {
	wm transient $w.entry [wm transient [winfo toplevel $w]]
    } else {
	wm transient $w.entry [winfo toplevel $w]
    }
    if [llength [wm group [winfo toplevel $w]]] {
	wm group $w.entry [wm group [winfo toplevel $w]]
    }
    wm title $w.entry "Entry box"
    wm resizable $w.entry 0 0
    catch {
	regexp {([-+][0-9]+)([-+][0-9]+)} [wm geometry [winfo toplevel $w]] geom geomX geomY
	wm geometry $w.entry +[expr 30 $geomX]+[expr 30 $geomY]
    }
    wm deiconify $w.entry

    bind $w.entry <Return> [list $w.entry.ok invoke]
    bind $w.entry <Escape> [list $w.entry.can invoke]

    with_grab_focus $w.entry $w.entry.e "tkwait window $w.entry"
    if $dirbrowser(dirlisteditentrybutton) {
	return -code error "Cancelled"
    }
    return $dirbrowser(dirlisteditentry)
}
;proc dirbrowser_edit_entry_but {w f} {
    global dirbrowser
    set dirbrowser(dirlisteditentrybutton) $f
    destroy $w
}

;proc dirbrowser_edit_up {w} {
    set sel [$w.l.l curselection]
    if {[llength $sel] && $sel>0} {
	set el [$w.l.l get $sel]
	$w.l.l delete $sel
	incr sel -1
	$w.l.l insert $sel $el
	$w.l.l selection set $sel
	$w.l.l activate $sel
	$w.l.l see $sel
    }
}
;proc dirbrowser_edit_down {w} {
    set sel [$w.l.l curselection]
    if {[llength $sel] && $sel<[$w.l.l index end]} {
	set el [$w.l.l get $sel]
	$w.l.l delete $sel
	incr sel
	$w.l.l insert $sel $el
	$w.l.l selection set $sel
	$w.l.l activate $sel
	$w.l.l see $sel
    }
}
;proc dirbrowser_edit_insert {w} {
    set sel [$w.l.l curselection]
    if [llength $sel] {
	incr sel
    } else {
	set sel 0
    }
    catch {
	set newel [dirbrowser_edit_entry $w.l.l [dirbrowser_pwd]]
	$w.l.l insert $sel $newel
	$w.l.l selection clear 0 end
	$w.l.l selection set $sel
	$w.l.l activate $sel
	$w.l.l see $sel
    }
}
;proc dirbrowser_edit_modify {w} {
    set sel [$w.l.l curselection]
    if [llength $sel] {
	catch {
	    set newel [dirbrowser_edit_entry $w.l.l [$w.l.l get $sel]]
	    $w.l.l delete $sel
	    $w.l.l insert $sel $newel
	    $w.l.l selection set $sel
	    $w.l.l activate $sel
	    $w.l.l see $sel
	}
    }
}
;proc dirbrowser_edit_delete {w} {
    set sel [$w.l.l curselection]
    if [llength $sel] {
	set last [$w.l.l index end]
	$w.l.l delete $sel
	if {$sel==$last} {
	    incr sel -1
	}
	$w.l.l selection set $sel
	$w.l.l activate $sel
	$w.l.l see $sel
    }
}
;proc dirbrowser_edit_ok {w} {
    global dirbrowser
    set dirbrowser(editbutton) 1
    set dirbrowser(editcontents) [$w.l.l get 0 end]
    destroy $w
}
;proc dirbrowser_edit_cancel {w} {
    global dirbrowser
    set dirbrowser(editbutton) 0
    destroy $w
}
;proc dirbrowser_edit_help {w} {
    tk_dialog $w.help "Directory List Editor Help" "Use the Up and Down buttons to move the currently\
	    selected directory through the list. Insert will create a new entry (below the current\
	    entry), Modify will change the current entry and Delete will remove the current entry." \
	    info 0 {O.K.}
}

# --- BEGINNING OF DEFAULT RESOURCES ---------------------------------------------------------------------

# This is a HACK. Prehaps I should badger JO into fixing tk_dialog so that it is possible to
# override the wrapLength... OK. I've mentioned this on comp.lang.tcl so this might get fixed
# inside the next couple of versions... :^(
catch {
    auto_load editor_help
}

option add  "*DKFDirbrowser.message.font"                "*Helvetica-Medium-R*18*"  startup
option add  "*DKFDirbrowser.filter.body.text.font"       "*Courier-Bold-R*12*"      startup
option add  "*DKFDirbrowser.filter.body.text.width"      55                         startup
option add  "*DKFDirbrowser.filename.body.text.font"     "*Courier-Bold-R*12*"      startup
option add  "*DKFDirbrowser.filename.body.text.width"    50                         startup
option add  "*DKFDirbrowser.dirs.body.2.list.font"       "*Helvetica-Bold-R*14*"    startup
option add  "*DKFDirbrowser.dirs.body.2.list.width"      25                         startup
option add  "*DKFDirbrowser.files.body.2.list.font"      "*Helvetica-Bold-R*14*"    startup
option add  "*DKFDirbrowser.files.body.2.list.width"     25                         startup
option add  "*DKFDirbrowser*Dialog*bitmap.foreground"    "blue"                     startup
option add  "*DKFDirBrowser*Dialog*msg.WrapLength"       5i                         startup
option add  "*DKFDirListEditor.l.l.width"                50                         startup
option add  "*DKFDirListEditor*Dialog*bitmap.foreground" "blue"                     startup
option add  "*DKFDirListEditorEntry.e.width"             50                         startup
